/************************************************************************
 *   IRC - Internet Relay Chat, ircd/channel.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Co Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "channel.h"
#include "msg.h" /* For TOK_*** and MSG_*** strings */
#include "h.h"
#include "unicode.h"
#include "send.h"

aChannel *channel = NullChn, *chlast = NullChn;

static void add_invite (aClient *, aChannel *);
static int add_banid (aClient *, aChannel *, char *);
static int add_excid (aClient *, aChannel *, char *);
static int add_invid (aClient *, aChannel *, char *);
static int can_join (aClient *, aChannel *, char *);
/*static*/ void channel_modes (aClient *, char *, char *, aChannel *);
int check_channelmask (aClient *, aClient *, char *);
static int del_banid (aChannel *, char *);
static int del_excid (aChannel *, char *);
static int del_invid (aChannel *, char *);
static int find_banid (aChannel *, char *);
static int find_excid (aChannel *, char *);
static int find_invid (aChannel *, char *);

static int have_ops (aChannel *);
static int set_mode (aClient *, aClient *, aChannel *, int, char **, char *, char *, int *, int);
static void sub1_from_channel (aChannel *);
aChannel *get_channel(aClient *, char *, int);

static int sendmodeto_one(aClient *, char *, char *, char *, char *, time_t);

extern int check_for_target_limit(aClient *, void *, const char *);

void clean_channelname (char *);
void del_invite (aClient *, aChannel *);

extern char *PartFmt, *PartFmt2;

/* some buffers for rebuilding channel/nick lists with ,'s */
static char nickbuf[BUFSIZE], buf[BUFSIZE];
static char modebuf[MODEBUFLEN], parabuf[MODEBUFLEN];

/* return the length (>=0) of a chain of links. */
static int list_length(Link * lp)
{
	int count = 0;

	for (; lp; lp = lp->next)
		count++;
	return count;
}

/*
** find_chasing
**     Find the client structure for a nick name (user) using history
**     mechanism if necessary. If the client is not found, an error
**     message (NO SUCH NICK) is generated. If the client was found
**     through the history, chasing will be 1 and otherwise 0.
*/
static aClient *find_chasing(aClient * sptr, char *user, int *chasing)
{
	aClient *who = find_client(user, (aClient *) NULL);

	if (chasing)
		*chasing = 0;
	if (who)
		return who;
	if ((who = get_history(user, (long) KILLCHASETIMELIMIT)) == NULL) {
		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, sptr->name, user);
		return NULL;
	}
	if (chasing)
		*chasing = 1;
	return who;
}

/*
 * Ban functions to work with mode +b
 * add_banid - add an id to be banned to the channel  (belongs to cptr)
 */

static int add_banid(aClient *cptr, aChannel *chptr, char *banid)
{
	Ban *ban;
	int cnt = 0, len = 0;

	if (MyClient(cptr))
		collapse(banid);

	for (ban = chptr->banlist; ban; ban = ban->next) {
		len += strlen(ban->banstr);
		if (MyClient(cptr))
			if ((len > MAXBANLENGTH) || (++cnt >= MAXBANS)) {
				sendto_one(cptr, err_str(ERR_BANLISTFULL), me.name, cptr->name, chptr->chname, banid);
				return -1;
			}
			else {
				if (!match(ban->banstr, banid) || !match(banid, ban->banstr))
					return -1;
			}
		else if (!mycmp(ban->banstr, banid))
			return -1;
	}

	ban = make_ban();
	memset(ban, 0, sizeof(Ban));
	ban->next = chptr->banlist;
	ban->banstr = (char *) MyMalloc(strlen(banid) + 1);
	strcpy(ban->banstr, banid);
	ban->who = (char *) MyMalloc(strlen(cptr->name) + 1);
	strcpy(ban->who, cptr->name);
	ban->when = time(NULL);
	chptr->banlist = ban;

	return 0;
}

/* Borg88
 * Exception functions to work with mode +e
 * add_excid - add an id to be except to the channel  (belongs to cptr)
 */

static int add_excid(aClient *cptr, aChannel *chptr, char *banid)
{
	Ban *ban;
	int cnt = 0, len = 0;

	if (MyClient(cptr))
		(void) collapse(banid);
	for (ban = chptr->exclist; ban; ban = ban->next) {
		len += strlen(ban->banstr);
		if (MyClient(cptr))
			if ((len > MAXBANLENGTH) || (++cnt >= MAXBANS)) {
				sendto_one(cptr, err_str(ERR_BANLISTFULL), me.name, cptr->name, chptr->chname, banid);
				return -1;
			}
			else {
				if (!match(ban->banstr, banid) || !match(banid, ban->banstr))
					return -1;
			}
		else if (!mycmp(ban->banstr, banid))
			return -1;

	}
	ban = make_ban();
	memset(ban, 0, sizeof(Ban));
	ban->next = chptr->exclist;
	ban->banstr = (char *) MyMalloc(strlen(banid) + 1);
	(void) strcpy(ban->banstr, banid);
	ban->who = (char *) MyMalloc(strlen(cptr->name) + 1);
	(void) strcpy(ban->who, cptr->name);
	ban->when = time(NULL);
	chptr->exclist = ban;
	return 0;
}

/* Borg88
 * Invite functions to work with mode +I
 * add_invid - add an id to be invited to the channel  (belongs to cptr)
 */

static int add_invid(cptr, chptr, banid)
aClient *cptr;
aChannel *chptr;
char *banid;
{
	Ban *ban;
	int cnt = 0, len = 0;

	if (MyClient(cptr))
		(void) collapse(banid);
	for (ban = chptr->invlist; ban; ban = ban->next) {
		len += strlen(ban->banstr);
		if (MyClient(cptr))
			if ((len > MAXBANLENGTH) || (++cnt >= MAXBANS)) {
				sendto_one(cptr, err_str(ERR_BANLISTFULL), me.name, cptr->name, chptr->chname, banid);
				return -1;
			}
			else {
				if (!match(ban->banstr, banid) || !match(banid, ban->banstr))
					return -1;
			}
		else if (!mycmp(ban->banstr, banid))
			return -1;

	}
	ban = make_ban();
	memset(ban, 0, sizeof(Ban));
	ban->next = chptr->invlist;
	ban->banstr = (char *) MyMalloc(strlen(banid) + 1);
	(void) strcpy(ban->banstr, banid);
	ban->who = (char *) MyMalloc(strlen(cptr->name) + 1);
	(void) strcpy(ban->who, cptr->name);
	ban->when = time(NULL);
	chptr->invlist = ban;
	return 0;
}

/*
 * del_banid - delete an id belonging to cptr
 */
static int del_banid(chptr, banid)
aChannel *chptr;
char *banid;
{
	Ban **ban;
	Ban *tmp;

	if (!banid)
		return -1;
	for (ban = &(chptr->banlist); *ban; ban = &((*ban)->next))
		if (mycmp(banid, (*ban)->banstr) == 0) {
			tmp = *ban;
			*ban = tmp->next;
			MyFree(tmp->banstr);
			MyFree(tmp->who);
			free_ban(tmp);
			return 0;
		}
	return -1;
}

/* Borg88
 * del_excid - delete an id belonging to cptr
 */
static int del_excid(aChannel * chptr, char *banid)
{
	Ban **ban;
	Ban *tmp;

	if (!banid)
		return -1;
	for (ban = &(chptr->exclist); *ban; ban = &((*ban)->next))
		if (mycmp(banid, (*ban)->banstr) == 0) {
			tmp = *ban;
			*ban = tmp->next;
			MyFree(tmp->banstr);
			MyFree(tmp->who);
			free_ban(tmp);
			return 0;
		}
	return -1;
}

/* del_invid - delete an id belonging to cptr --  Borg88 */
static int del_invid(aChannel * chptr, char *banid)
{
	Ban **ban;
	Ban *tmp;

	if (!banid)
		return -1;
	for (ban = &(chptr->invlist); *ban; ban = &((*ban)->next))
		if (mycmp(banid, (*ban)->banstr) == 0) {
			tmp = *ban;
			*ban = tmp->next;
			MyFree(tmp->banstr);
			MyFree(tmp->who);
			free_ban(tmp);
			return 0;
		}
	return -1;
}

/* find_banid - Find an exact match for a ban */
static int find_banid(aChannel * chptr, char *banid)
{
	Ban **ban;

	if (!banid)
		return -1;
	for (ban = &(chptr->banlist); *ban; ban = &((*ban)->next))
		if (!mycmp(banid, (*ban)->banstr))
			return 1;
	return 0;
}

/* find_excid - Find an exact match for an Exception --  Borg88 */
static int find_excid(aChannel * chptr, char *banid)
{
	Ban **ban;

	if (!banid)
		return -1;
	for (ban = &(chptr->exclist); *ban; ban = &((*ban)->next))
		if (!mycmp(banid, (*ban)->banstr))
			return 1;
	return 0;
}

/* Borg88
 * find_invid - Find an exact match for an Invite
 */
static int find_invid(aChannel * chptr, char *banid)
{
	Ban **ban;

	if (!banid)
		return -1;
	for (ban = &(chptr->invlist); *ban; ban = &((*ban)->next))
		if (!mycmp(banid, (*ban)->banstr))
			return 1;
	return 0;
}


/* returns 1 if a person is joined and not a zombie */
int IsMember(aClient * cptr, aChannel * chptr)
{
	return (find_user_link(chptr->members, cptr) != NULL) ? 1 : 0;
}


/* is_banned - returns a pointer to the ban structure if banned else NULL */
Ban *is_banned(aClient *cptr, aChannel *chptr, int skipfwd)
{
	char *comma = (char*)NULL;
	Ban *ban, *exc;
	char host[NICKLEN + USERLEN + HOSTLEN + 6],
		vhost[NICKLEN + USERLEN + HOSTLEN + 6];

	if (!IsPerson(cptr))
		return 0;

	make_nick_user_hosts(cptr, host, vhost);

	for (ban = chptr->banlist; ban; ban = ban->next) {
		if ((comma = strchr(ban->banstr, ',')) != NULL) {
			if (skipfwd)
				continue;
			*comma = '\0';
		}

		if (!match(ban->banstr, host) || !match(ban->banstr, vhost))
			break;

		if (comma)
			*comma = ',';
	}

	if (comma)
		*comma = ',';

	/* not banned anyways */
	if (ban) {
		for (exc = chptr->exclist; exc; exc = exc->next) {
			if (!match(exc->banstr, host) || !match(exc->banstr, vhost)) {
				ban = (Ban*) NULL;
				break;
			}
		}
	}

	return ban;
}


/* checks if client is on the invite list */
static int is_oninv(aClient *cptr, aChannel *chptr)
{
	Ban *tmp;
	char host[NICKLEN + USERLEN + HOSTLEN + 6],
		vhost[NICKLEN + USERLEN + HOSTLEN + 6];

	if (!IsPerson(cptr))
		return 0;

	make_nick_user_hosts(cptr, host, vhost);

	for (tmp = chptr->invlist; tmp; tmp = tmp->next) {
		if (!match(tmp->banstr, host) || !match(tmp->banstr, vhost))
			return 1;
	}

	return 0;
}


/*
 * adds a user to a channel by adding another link to the channels member
 * chain.
 */
static void add_user_to_channel(aChannel * chptr, aClient * who, int flags)
{
	struct SLink *ptr;

	if (who->user) {
		ptr = make_link();
		ptr->value.cptr = who;
		ptr->flags = flags;
		ptr->next = chptr->members;
		chptr->members = ptr;
		chptr->users++;

		ptr = make_link();
		ptr->value.chptr = chptr;
		ptr->next = who->user->channel;
		who->user->channel = ptr;
		who->user->joined++;
	}
}

void remove_user_from_channel(aClient * sptr, aChannel * chptr)
{
	Link **curr, *tmp, *lp = chptr->members;

	for (; lp && (lp->flags & CHFL_ZOMBIE || lp->value.cptr == sptr); lp = lp->next);

	for (;;) {
		for (curr = &chptr->members; (tmp = *curr) != NULL; curr = &tmp->next)
			if (tmp->value.cptr == sptr) {
				*curr = tmp->next;
				free_link(tmp);
				break;
			}

		for (curr = &sptr->user->channel; (tmp = *curr) != NULL; curr = &tmp->next)
			if (tmp->value.chptr == chptr) {
				*curr = tmp->next;
				free_link(tmp);
				break;
			}

		--(sptr->user->joined);

		if (lp)
			break;

		if (chptr->members)
			sptr = chptr->members->value.cptr;
		else
			break;

		sub1_from_channel(chptr);
	}
	sub1_from_channel(chptr);
}

static int have_ops(aChannel * chptr)
{
	struct SLink *lp;

	if (chptr) {
		lp = chptr->members;
		while (lp) {
			if (lp->flags & CHFL_CHANOP)
				return (1);
			lp = lp->next;
		}
	}
	return 0;
}

int has_op(aClient *cptr, aChannel *chptr)
{
	struct SLink *lp;

	return (chptr && (lp = find_user_link(chptr->members, cptr)) != NULL)
		? (lp->flags & CHFL_CHANOP)
		: 0;
}

int is_chan_op(aClient * cptr, aChannel * chptr)
{
	struct SLink *lp;

	return (chptr && (lp = find_user_link(chptr->members, cptr)) != NULL)
		? lp->flags & CHFL_CHANOP : 0;
}


/* strip colour codes from +C channels */
int strip_color(char *message)
{
	int count;
	char *dst, c;

	for (count = 0, dst = message; (c = *message) != 0; ) {
		/* strip reversed text also */
		if (c == '\x16') {
			++count, ++message;
			continue;
		}

		if (c != '\x03') {
			*dst++ = c;
			++message;
		}
		else {
			++count, ++message;

			if (isdigit(*message))
				++message;
			if (isdigit(*message))
				++message;

			if (*message != ',')
				continue;

			if (isdigit(*++message))
				++message;
			if (isdigit(*message))
				++message;
		}
	}

	*dst = '\0';

	return count;
}


/* Rewrited. Evil TS=0 fixed. -- Borg88 */

/* parv[0] - sender, parv[1] - channel */
int m_mode(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	static char tmp[MODEBUFLEN];
	int badop, sendts;
	aChannel *chptr;

	if (check_registered(sptr))
		return 0;

	/* Now, try to find the channel in question */
	if (parc < 2) {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "MODE");
		return 0;
	}

	/* nonchannels are users */
	if (!IsChannelName(parv[1]))
		return m_umode(cptr, sptr, parc, parv);

	if (IsModelessChannel(parv[1])) {
		sendto_one(sptr, ":%s 403 %s %s :Channel does not support modes", me.name, parv[0], parv[1]);
		return 0;
	}

	if ((chptr = find_channel(parv[1], NullChn)) == NullChn && !IsServer(sptr) && !IsULine(sptr, cptr)) {
		sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], parv[1]);
		return 0;
	}

	sptr->flags &= ~FLAGS_TS8;

	if (MyConnect(sptr))
		clean_channelname(parv[1]);

	if (check_channelmask(sptr, cptr, parv[1]))
		return 0;

	if (parc < 3) {
		*modebuf = 0;
		*parabuf = 0;
		modebuf[1] = 0;
		channel_modes(sptr, modebuf, parabuf, chptr);
		sendto_one(sptr, rpl_str(RPL_CHANNELMODEIS), me.name, parv[0], chptr->chname, modebuf, parabuf);
		sendto_one(sptr, rpl_str(RPL_CREATIONTIME), me.name, parv[0], chptr->chname, chptr->creationtime);
		return 0;
	}

	/* if the channel does not exist - create it, will get killed below if
	   not persistent.  not nice, but the privileged clients should not send
	   that much of modes for unexisting channels. -- hex */
	chptr = get_channel(sptr, parv[1], CREATE);

	if (!(sendts = set_mode(cptr, sptr, chptr, parc - 2, parv + 2, modebuf, parabuf, &badop, 0)) && !IsULine(cptr, sptr)) {
		sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0], chptr->chname);

		/* kill empty dynamic channels */
		if (chptr->users <= 0 && !PersistantChannel(chptr))
			sub1_from_channel(chptr);

		return 0;
	}

	if ((badop >= 2) && !IsULine(cptr, sptr)) {
		int i = 3;

		*tmp = 0;
		while (i < parc) {
			strcat(tmp, " ");
			strcat(tmp, parv[i++]);
		}
	}

	if (strlen(modebuf) > (size_t) 1 || sendts > 0) {
		if (badop != 2 && strlen(modebuf) > (size_t) 1)
			sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", parv[0], chptr->chname, modebuf, parabuf);

		/* We send a creationtime of 0, to mark it as a hack --Run */
		if ((IsServer(sptr) && (badop == 2 || sendts > 0)) || IsULine(cptr, sptr)) {
			if (*modebuf == 0)
				strcpy(modebuf, "+");
			if (badop != 2)
				sendto_match_servs(chptr, cptr, ":%s MODE %s %s %s %lu", parv[0], chptr->chname, modebuf, parabuf, (badop == 4) ? (time_t) 0 : chptr->creationtime);
		}
		else {
			sendto_match_servs(chptr, cptr, ":%s MODE %s %s %s", parv[0], chptr->chname, modebuf, parabuf);
		}
	}

	/* kill empty dynamic channels here too */
	if (chptr->users <= 0 && !PersistantChannel(chptr))
		sub1_from_channel(chptr);

	return 0;
}

int has_voice(aClient * cptr, aChannel * chptr)
{
	struct SLink *lp;

	if (chptr)
		if ((lp = find_user_link(chptr->members, cptr)) != NULL && !(lp->flags & CHFL_ZOMBIE))
			return (lp->flags & CHFL_VOICE);
	return 0;
}


const char * can_send(aClient *cptr, aChannel *chptr, const char *msg)
{
	Link *lp;
	int member;

	if (IsULine(cptr, cptr) || IsServer(cptr))
		return 0;

	member = IsMember(cptr, chptr);
	lp = find_user_link(chptr->members, cptr);

	if (chptr->mode.mode & MODE_MODERATED && (!lp || !((lp->flags & (CHFL_CHANOP | CHFL_VOICE))) || (lp->flags & CHFL_ZOMBIE)))
		return "moderated";

	if (chptr->mode.mode & MODE_NOPRIVMSGS && !member)
		return "messages from outside are disallowed";

	if ((!lp || !((lp->flags & (CHFL_CHANOP | CHFL_VOICE))) || (lp->flags & CHFL_ZOMBIE)) && MyClient(cptr) && is_banned(cptr, chptr, 1))
		return "banned";

	if ((chptr->mode.mode & MODE_NOCOLOR) && MyClient(cptr) && (!member || !(lp->flags & CHFL_CHANOP))) {
		while (*msg != '\0') {
			if (*msg == 0x03 || *msg == 0x16)
				return "colors are forbidden";
			++msg;
		}
	}

	return 0;
}

aChannel *find_channel(char *chname, aChannel * chptr)
{
	return hash_find_channel(chname, chptr);
}

/*
 * write the "simple" list of channel modes for channel chptr onto buffer mbuf
 * with the parameters in pbuf.
 */
void channel_modes(aClient * cptr, char *mbuf, char *pbuf, aChannel * chptr)
{
	*mbuf++ = '+';
	if (chptr->mode.mode & MODE_SECRET)
		*mbuf++ = 's';
	else if (chptr->mode.mode & MODE_PRIVATE)
		*mbuf++ = 'p';
	if (chptr->mode.mode & MODE_FORWARD)
		*mbuf++ = 'f';
	if (chptr->mode.mode & MODE_JUPED)
		*mbuf++ = 'j';
	if (chptr->mode.mode & MODE_MODERATED)
		*mbuf++ = 'm';
	if (chptr->mode.mode & MODE_TOPICLIMIT)
		*mbuf++ = 't';
	if (chptr->mode.mode & MODE_INVITEONLY)
		*mbuf++ = 'i';
	if (chptr->mode.mode & MODE_NOPRIVMSGS)
		*mbuf++ = 'n';
	if (chptr->mode.mode & MODE_RGSTR)
		*mbuf++ = 'r';
	if (chptr->mode.mode & MODE_RGSTRONLY)
		*mbuf++ = 'R';
	if (chptr->mode.mode & MODE_NOCOLOR)
		*mbuf++ = 'c';
	if (chptr->mode.limit) {
		*mbuf++ = 'l';
		if (pbuf && (IsMember(cptr, chptr) || IsServer(cptr)
		|| IsULine(cptr, cptr)))
			sprintf(pbuf, "%d ", chptr->mode.limit);
	}
	if (*chptr->mode.key) {
		*mbuf++ = 'k';
		if (pbuf && (IsMember(cptr, chptr) || IsServer(cptr)
		|| IsULine(cptr, cptr)))
			strcat(pbuf, chptr->mode.key);
	}

	*mbuf = '\0';
	return;
}

static int send_mode_list(aClient *cptr, char *chname, time_t creationtime, Link *top, int mask, char flag)
{
	Link *lp;
	char *cp, *name;
	int count = 0, send = 0, sent = 0;

	cp = modebuf + strlen(modebuf);
	if (*parabuf) /* mode +l or +k xx */
		count = 1;
	for (lp = top; lp; lp = lp->next) {
		/*
		 * Okay, since ban's are stored in their own linked
		 * list, we won't even bother to check if CHFL_BAN
		 * is set in the flags. This should work as long
		 * as only ban-lists are feed in with CHFL_BAN mask.
		 * However, we still need to typecast... -Donwulff
		 */
		if (mask == CHFL_BAN) {
			name = ((Ban *) lp)->banstr;
		}
		else {
			if (!(lp->flags & mask))
				continue;
			name = lp->value.cptr->name;
		}
		if (strlen(parabuf) + strlen(name) + 11 < (size_t) MODEBUFLEN) {
			if (*parabuf)
				(void) strcat(parabuf, " ");
			(void) strcat(parabuf, name);
			count++;
			*cp++ = flag;
			*cp = '\0';
		}
		else if (*parabuf)
			send = 1;
		if (count == RESYNCMODES)
			send = 1;
		if (send) {
			/* cptr is always a server! So we send creationtimes */
			sendmodeto_one(cptr, me.name, chname, modebuf, parabuf, creationtime);
			sent = 1;
			send = 0;
			*parabuf = '\0';
			cp = modebuf;
			*cp++ = '+';
			if (count != RESYNCMODES) {
				(void) strcpy(parabuf, name);
				*cp++ = flag;
			}
			count = 0;
			*cp = '\0';
		}
	}
	return sent;
}

/*
 * send "cptr" a full list of the modes for channel chptr.
 */
void send_channel_modes(cptr, chptr)
aClient *cptr;
aChannel *chptr;
{
	int sent;

	if (*chptr->chname != '#')
		return;

	*modebuf = *parabuf = '\0';
	channel_modes(cptr, modebuf, parabuf, chptr);

	sent = send_mode_list(cptr, chptr->chname, chptr->creationtime, chptr->members, CHFL_CHANOP, 'o');
	if (!sent && chptr->creationtime)
		sendto_one(cptr, ":%s %s %s %s %s %lu", me.name, (IsToken(cptr) ? TOK_MODE : MSG_MODE), chptr->chname, modebuf, parabuf, chptr->creationtime);
	else if (modebuf[1] || *parabuf)
		sendmodeto_one(cptr, me.name, chptr->chname, modebuf, parabuf, chptr->creationtime);

	*parabuf = '\0';
	*modebuf = '+';
	modebuf[1] = '\0';
	(void) send_mode_list(cptr, chptr->chname, chptr->creationtime, (Link*)chptr->banlist, CHFL_BAN, 'b');
	(void) send_mode_list(cptr, chptr->chname, chptr->creationtime, (Link*)chptr->exclist, CHFL_BAN, 'e');
	(void) send_mode_list(cptr, chptr->chname, chptr->creationtime, (Link*)chptr->invlist, CHFL_BAN, 'I');
	if (modebuf[1] || *parabuf)
		sendmodeto_one(cptr, me.name, chptr->chname, modebuf, parabuf, chptr->creationtime);

	*parabuf = '\0';
	*modebuf = '+';
	modebuf[1] = '\0';
	(void) send_mode_list(cptr, chptr->chname, chptr->creationtime, chptr->members, CHFL_VOICE, 'v');
	if (modebuf[1] || *parabuf)
		sendmodeto_one(cptr, me.name, chptr->chname, modebuf, parabuf, chptr->creationtime);
}

/* server mode for opers -- Borg88 */
int m_smode(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	aChannel *chptr;
	int badop;

	if (check_registered(sptr))
		return 0;

	if (!IsOper(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	if (parc < 3) {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SMODE");
		return 0;
	}

	if ((chptr = find_channel(parv[1], NullChn)) == NULL) {
		sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], parv[1]);
		return 0;
	}

	if (IsModelessChannel(parv[1])) {
		sendto_one(sptr, ":%s 403 %s %s :Channel does not support modes", me.name, parv[0], parv[1]);
		return 0;
	}

	sptr->flags &= ~FLAGS_TS8;
	*modebuf = 0;
	*parabuf = 0;

	if (!set_mode(cptr, sptr, chptr, parc - 2, parv + 2, modebuf, parabuf, &badop, 1)) {
		sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0], chptr->chname);
		return 0;
	}

	if (*modebuf) {
		sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", me.name, chptr->chname, modebuf, parabuf);
		sendto_match_servs(chptr, cptr, ":%s MODE %s %s %s 0", me.name, chptr->chname, modebuf, parabuf);
	}

	/* kill empty non-persistant channels */
	if (chptr->users <= 0 && !PersistantChannel(chptr))
		sub1_from_channel(chptr);

	return 0;
}

int DoesOp(char *modebuf)
{
	modebuf--; /* Is it possible that a mode starts with o and not +o ? */
	while (*++modebuf)
		if (*modebuf == 'o' || *modebuf == 'v')
			return (1);
	return 0;
}

int sendmodeto_one(aClient *cptr, char *from, char *name,
	char *mode, char *param, time_t creationtime)
{
	if ((IsServer(cptr) && DoesOp(mode) && creationtime) || IsULine(cptr, cptr))
		sendto_one(cptr, ":%s %s %s %s %s %lu", from, (IsToken(cptr) ? TOK_MODE : MSG_MODE), name, mode, param, creationtime);
	else
		sendto_one(cptr, ":%s %s %s %s %s", from, (IsToken(cptr) ? TOK_MODE : MSG_MODE), name, mode, param);
	return 0;
}

char *pretty_mask(char *mask)
{
	char *p, *fwd;
	int flags = 0; /* nick = 1, user = 2, dots = 4, wild = 8 */
	static char pretty[NICKLEN + USERLEN + HOSTLEN + CHANNELLEN + 7];

	strip_color(mask);
	collapse(mask);

	/* find forwarding mask and truncate there */
	if ((fwd = strchr(mask, ',')) != NULL)
		*fwd = '\0';

	/* copy it to offset 3 since we may later add up to 3 characters */
	strcpy(pretty + 3, mask);

	/* scan the mask and truncate if anything bad happens */
	for (p = pretty + 3; *p; ++p) {
		if (*p == '!') {
			if (flags & 7) { /* can't have an @ after ! or a dot */
				*p = '\0';
				break;
			}
			flags |= 1;
		}
		else if (*p == '@') { /* this can follow both ! and a dot */
			if (flags & 2) {
				*p = '\0';
				break;
			}
			flags |= 2;
		}
		else if (*p == '.')
			flags |= 4;
		else if (*p == '*')
			flags |= 8;
	}

	/* it's where our mask starts now */
	p = pretty + 3;

	/* a plain nickname */
	if (!flags) {
		strcat(p, "!*@*");
		flags = 15;
	}

	/* nick!user, an ! with no @ */
	if ((flags & 3) == 1) {
		switch (p[strlen(p) - 1]) {
			case '!':
				strcat(p, "*@*");
				break;
			case '*': /* nick@user* -- it's okay already */
				break;
			default:
				strcat(p, "@*");
				break;
		}
		flags = 15;
	}

	/* user@host -- add empty nick if not an open mask */
	if ((flags & 3) == 2 && *p != '*') {
		*--p = '!';
		*--p = '*';
		flags |= 1;
	}

	/* ain't got neither ! nor @, but it's a hostname (got [.*]) */
	if (!(flags & (1+2+6))) {
		if (*p == '.')
			*--p = '*';
		if (*p != '*') {
			*--p = '@';
			*--p = '*';
		}
	}

	/* empty nickname */
	if (*p == '!')
		*--p = '*';

	/* append target of forwarding */
	if (fwd) {
		strcat(p, ",");
		strcat(p, fwd + 1);
	}

	return p;
}

/* Check and try to apply the channel modes passed in the parv array for
 * the client ccptr to channel chptr.  The resultant changes are printed
 * into mbuf and pbuf (if any) and applied to the channel. */
static int set_mode(aClient * cptr, aClient * sptr, aChannel * chptr, int parc, char *parv[], char *mbuf, char *pbuf, int *badop, int sadmin)
{
	static Link chops[MAXMODEPARAMS];
	static int flags[] = {
		MODE_INVITEONLY, 'i',
		MODE_JUPED, 'j',
		MODE_KEY, 'k',
		MODE_MODERATED, 'm',
		MODE_NOPRIVMSGS, 'n',
		MODE_PRIVATE, 'p',
		MODE_RGSTR, 'r',
		MODE_SECRET, 's',
		MODE_TOPICLIMIT, 't',
		MODE_VOICE, 'v',
		MODE_NOCOLOR, 'c',
		MODE_FORWARD, 'f',
		MODE_RGSTRONLY, 'R',
		0, 0
	};
	Ban *ban;
	Link *lp;
	char *curr = parv[0], *cp = NULL;
	int *ip;
	Link *member, *tmp = NULL;
	u_int whatt = MODE_ADD, bwhatt = 0;
	int limitset = 0, chasing = 0, bounce, nusers = 0, new, len, blen,
		keychange = 0, opcnt = 0, banlsent = 0, exclsent = 0,
		invlsent = 0 /*, fwdset = 0*/;
	int doesdeop = 0, doesop = 0, hacknotice = 0, change, gotts = 0;
	/* char fm = 0; */
	aClient *who;
	Mode *mode, oldm;
	static char bmodebuf[MODEBUFLEN], bparambuf[MODEBUFLEN], numeric[16];
	char *bmbuf = bmodebuf, *bpbuf = bparambuf /*, *mbufp = mbuf*/;
	time_t newtime = (time_t) 0;
	aConfItem *aconf;

	*mbuf = 0;
	*pbuf = 0;
	*bmbuf = 0;
	*bpbuf = 0;
	*badop = 0;

	if (parc < 1)
		return 0;

	mode = &(chptr->mode);
	memcpy(&oldm, mode, sizeof(Mode));

/* Mode is accepted when sptr is a channel operator
 * but also when the mode is received from a server. --Run */

/* This was modified to allow non-chanops to see bans..perhaps not as clean
 * as other versions, but it works. Basically, if a person who's not a chanop
 * calls set_mode() with a /mode <some parameter here> command, this function
 * checks to see if that parameter is +b, -b, or just b. If it is one of
 * these, and there are NO other parameters (i.e. no multiple mode changes),
 * then it'll display the banlist. Else, it just returns "You're not chanop."
 * Enjoy. --dalvenjah, dalvenja@rahul.net */

	if (!(IsServer(cptr) || IsULine(cptr, sptr) || is_chan_op(sptr, chptr) || sadmin == 1)) {
		if (((*curr == 'b' && (strlen(parv[0]) == 1 && parc == 1))
			 || ((*curr == '+') && (*(curr + 1) == 'b') && (strlen(parv[0]) == 2 && parc == 1))
			 || ((*curr == '-') && (*(curr + 1) == 'b') && (strlen(parv[0]) == 2 && parc == 1)))
			&& (!is_chan_op(sptr, chptr))) {
			if ((int)--parv <= 0) {
				for (ban = chptr->banlist; ban; ban = ban->next)
					sendto_one(cptr, rpl_str(RPL_BANLIST), me.name, cptr->name, chptr->chname, ban->banstr, ban->who, ban->when);
				sendto_one(cptr, rpl_str(RPL_ENDOFBANLIST), me.name, cptr->name, chptr->chname);
			}
		}
		else if (((*curr == 'e' && (strlen(parv[0]) == 1 && parc == 1))
				  || ((*curr == '+') && (*(curr + 1) == 'e') && (strlen(parv[0]) == 2 && parc == 1))
				  || ((*curr == '-') && (*(curr + 1) == 'e') && (strlen(parv[0]) == 2 && parc == 1)))
				 && (!is_chan_op(sptr, chptr))) {
			if ((int)--parv <= 0) {
				for (ban = chptr->exclist; ban; ban = ban->next)
					sendto_one(cptr, rpl_str(RPL_EXCLIST), me.name, cptr->name, chptr->chname, ban->banstr, ban->who, ban->when);
				sendto_one(cptr, rpl_str(RPL_ENDOFEXCLIST), me.name, cptr->name, chptr->chname);
			}
		}
		else if (((*curr == 'I' && (strlen(parv[0]) == 1 && parc == 1))
				  || ((*curr == '+') && (*(curr + 1) == 'I') && (strlen(parv[0]) == 2 && parc == 1))
				  || ((*curr == '-') && (*(curr + 1) == 'I') && (strlen(parv[0]) == 2 && parc == 1)))
				 && (!is_chan_op(sptr, chptr))) {
			if ((int)--parv <= 0) {
				for (ban = chptr->invlist; ban; ban = ban->next)
					sendto_one(cptr, rpl_str(RPL_INVLIST), me.name, cptr->name, chptr->chname, ban->banstr, ban->who, ban->when);
				sendto_one(cptr, rpl_str(RPL_ENDOFINVLIST), me.name, cptr->name, chptr->chname);
			}
		}
		else
			return 0;
	}

	new = mode->mode;
	while (curr && *curr) {
		switch (*curr) {
		case '+':
			whatt = MODE_ADD;
			break;

		case '-':
			whatt = MODE_DEL;
			break;

		case 'o':
		case 'v':
			if (--parc <= 0)
				break;
			parv++;
			*parv = check_string(*parv);
			if (opcnt >= MODEPARAMS)
				break;
			/* Check for nickname changes and try to follow these
			 * to make sure the right client is affected by the
			 * mode change. */
			if ((who = find_chasing(sptr, parv[0], &chasing)) == NULL)
				break;

			/* don't allow to deop/devoice services */
			if (IsULine(who, who) && whatt == MODE_DEL) {
				sendto_one(cptr, ":%s NOTICE %s :Cannot de%s service", me.name, cptr->name, *curr == 'o' ? "op" : "voice");
				break;
			}

			if ((member = find_user_link(chptr->members, who)) == NULL) {
				sendto_one(cptr, err_str(ERR_USERNOTINCHANNEL), me.name, cptr->name, parv[0], chptr->chname);
				break;
			}
			if (whatt == MODE_ADD) {
				lp = &chops[opcnt++];
				lp->value.cptr = who;
				if ((IsServer(sptr) && (!(who->flags & FLAGS_TS8) || ((*curr == 'o') && !(member->flags & (CHFL_SERVOPOK | CHFL_CHANOP))) || who->from != sptr->from)) || IsULine(cptr, sptr) || sadmin == 1) {
					*badop = ((member->flags & CHFL_DEOPPED) && (*curr == 'o')) ? 2 : 3;
				}
				lp->flags = (*curr == 'o') ? MODE_CHANOP : MODE_VOICE;
				lp->flags |= MODE_ADD;
			}
			else if (whatt == MODE_DEL) {
				lp = &chops[opcnt++];
				lp->value.cptr = who;
				doesdeop = 1; /* Also when -v */
				lp->flags = (*curr == 'o') ? MODE_CHANOP : MODE_VOICE;
				lp->flags |= MODE_DEL;
			}
			if (*curr == 'o')
				doesop = 1;
			break;

		case 'k':
			if (--parc <= 0)
				break;
			parv++;
			/* check now so we eat the parameter if present */
			if (keychange)
				break;
			*parv = check_string(*parv);
			{
				u_char *s1, *s2;

				for (s1 = s2 = (u_char *) * parv; *s2; s2++) {
					if ((u_char)(*s1 = (u_char)(*s2 & 0x7f)) > (u_char) 32 && *s1 != ':')
						s1++;
				}
				*s1 = 0;
			}

			if (opcnt >= MODEPARAMS)
				break;

			if (whatt == MODE_ADD) {
				if (*mode->key && !IsServer(cptr) && !IsULine(cptr, sptr) && sadmin == 0)
					sendto_one(cptr, err_str(ERR_KEYSET), me.name, cptr->name, chptr->chname);
				else if (!*mode->key || IsServer(cptr) || IsULine(cptr, sptr) || sadmin == 1) {
					lp = &chops[opcnt++];
					lp->value.cp = *parv;
					if (strlen(lp->value.cp) > (size_t) KEYLEN)
						lp->value.cp[KEYLEN] = 0;
					lp->flags = MODE_KEY | MODE_ADD;
					keychange = 1;
				}
			}
			else if (whatt == MODE_DEL) {
				if (mycmp(mode->key, *parv) == 0 || IsServer(cptr) || IsULine(cptr, sptr) || sadmin == 1) {
					lp = &chops[opcnt++];
					lp->value.cp = mode->key;
					lp->flags = MODE_KEY | MODE_DEL;
					keychange = 1;
				}
			}
			break;

		case 'b':
			if (--parc <= 0) {
				if (banlsent) /* Only send it once */
					break;
				for (ban = chptr->banlist; ban; ban = ban->next)
					sendto_one(cptr, rpl_str(RPL_BANLIST), me.name, cptr->name, chptr->chname, ban->banstr, ban->who, ban->when);
				sendto_one(cptr, rpl_str(RPL_ENDOFBANLIST), me.name, cptr->name, chptr->chname);
				banlsent = 1;
				break;
			}

			parv++;
			if (BadPtr(*parv))
				break;

			if (opcnt >= MODEPARAMS)
				break;

			if (whatt == MODE_ADD) {
				lp = &chops[opcnt++];
				lp->value.cp = *parv;
				lp->flags = MODE_ADD | MODE_BAN;
			}
			else if (whatt == MODE_DEL) {
				lp = &chops[opcnt++];
				lp->value.cp = *parv;
				lp->flags = MODE_DEL | MODE_BAN;
			}

			break;

		case 'e':
			if (--parc <= 0) {
				if (exclsent) /* Only send it once */
					break;
				for (ban = chptr->exclist; ban; ban = ban->next)
					sendto_one(cptr, rpl_str(RPL_EXCLIST), me.name, cptr->name, chptr->chname, ban->banstr, ban->who, ban->when);
				sendto_one(cptr, rpl_str(RPL_ENDOFEXCLIST), me.name, cptr->name, chptr->chname);
				exclsent = 1;
				break;
			}
			parv++;
			if (BadPtr(*parv))
				break;
			if (opcnt >= MODEPARAMS)
				break;
			if (whatt == MODE_ADD) {
				lp = &chops[opcnt++];
				lp->value.cp = *parv;
				lp->flags = MODE_ADD | MODE_EXC;
			}
			else if (whatt == MODE_DEL) {
				lp = &chops[opcnt++];
				lp->value.cp = *parv;
				lp->flags = MODE_DEL | MODE_EXC;
			}
			break;

		case 'I':
			if (--parc <= 0) {
				if (invlsent) /* Only send it once */
					break;
				for (ban = chptr->invlist; ban; ban = ban->next)
					sendto_one(cptr, rpl_str(RPL_INVLIST), me.name, cptr->name, chptr->chname, ban->banstr, ban->who, ban->when);
				sendto_one(cptr, rpl_str(RPL_ENDOFINVLIST), me.name, cptr->name, chptr->chname);
				invlsent = 1;
				break;
			}
			parv++;
			if (BadPtr(*parv))
				break;
			if (opcnt >= MODEPARAMS)
				break;
			if (whatt == MODE_ADD) {
				lp = &chops[opcnt++];
				lp->value.cp = *parv;
				lp->flags = MODE_ADD | MODE_INV;
			}
			else if (whatt == MODE_DEL) {
				lp = &chops[opcnt++];
				lp->value.cp = *parv;
				lp->flags = MODE_DEL | MODE_INV;
			}
			break;

		case 'l':
			/* limit 'l' to only *1* change per mode command but eat up others. */
			if (limitset) {
				if (whatt == MODE_ADD && --parc > 0)
					parv++;
				break;
			}
			if (whatt == MODE_DEL) {
				limitset = 1;
				nusers = 0;
				break;
			}
			if (--parc > 0) {
				if (BadPtr(*parv))
					break;
				if (opcnt >= MODEPARAMS)
					break;
				if (!(nusers = atoi(*++parv)))
					continue;
				lp = &chops[opcnt++];
				lp->flags = MODE_ADD | MODE_LIMIT;
				limitset = 1;
				break;
			}
			sendto_one(cptr, err_str(ERR_NEEDMOREPARAMS), me.name, cptr->name, "MODE +l");
			break;

		case 'f':
			if (!IsPrivileged(cptr) || IsLocOp(cptr)) {
				sendto_one(cptr, err_str(ERR_ONLYSERVERSCANCHANGE),
					me.name, cptr->name, chptr->chname,
					"IRC Operators", *curr);
				break;
			}
			if (whatt == MODE_DEL)
				new &= ~MODE_FORWARD;
			else
				new |= MODE_FORWARD;
			break;

		case 'j':
			if (!IsPrivileged(cptr)) {
				sendto_one(cptr, err_str(ERR_ONLYSERVERSCANCHANGE),
					me.name, cptr->name, chptr->chname,
					"IRC Operators", *curr);
				break;
			}

			if (whatt == MODE_DEL)
				new &= ~MODE_JUPED;
			else
				new |= MODE_JUPED;

			break;

		case 'r':
			if (!IsServer(cptr)) {
				sendto_one(cptr, err_str(ERR_ONLYSERVERSCANCHANGE), me.name, cptr->name, chptr->chname, "servers", *curr);
				break;
			}
			if (whatt == MODE_DEL)
				new &= ~MODE_RGSTR;
			else
				new |= MODE_RGSTR;
			break;

		/* clean the topic here, instead of on every TOPIC command */
		case 'c':
			if (whatt == MODE_ADD) {
				new |= MODE_NOCOLOR;
				if (chptr->topic)
					strip_color(chptr->topic);
			}

		case 'i': /* falls through for default case */
			if (whatt == MODE_DEL) {
				while ((lp = chptr->invites) != NULL)
					del_invite(lp->value.cptr, chptr);
			}

		default:
			for (ip = flags; *ip; ip += 2) {
				if (*(ip + 1) == *curr)
					break;
			}
			if (*ip) {
				if (whatt == MODE_ADD) {
					if (*ip == MODE_PRIVATE)
						new &= ~MODE_SECRET;
					else if (*ip == MODE_SECRET)
						new &= ~MODE_PRIVATE;
					new |= *ip;
				}
				else
					new &= ~*ip;
			}
			else if (!IsServer(cptr) && !IsULine(cptr, sptr))
				sendto_one(cptr, err_str(ERR_UNKNOWNMODE), me.name, cptr->name, *curr, chptr->chname);
			break;
		}

		curr++;

		/* Make sure mode strings such as "+m +t +p +i" are parsed *fully. */
		/* but only from servers -- hex */
		if ((IsServer(sptr) || IsULine(cptr, sptr)) && !*curr && parc > 0) {
			curr = *++parv;
			parc--;
			/* If this was from a server, and it is the last parameter and it
			   starts with a digit, it must be the creationtime.  --Run */
			if (IsServer(sptr) || IsULine(cptr, sptr)) {
				if (parc == 1 && isdigit(*curr)) {
					newtime = atoi(curr);
					gotts = 1;
					if (newtime == 0) {
						*badop = 2;
						hacknotice = 1;
					}
					else if (newtime > chptr->creationtime) {
						/* It is a net-break ride if we have ops. bounce modes
						   if we have ops. --Run */
						if (doesdeop)
							*badop = 2;
						else if (chptr->creationtime == 0 || (!have_ops(chptr) && !PersistantChannel(chptr))) {
							if (chptr->creationtime == 0 || doesop)
								chptr->creationtime = newtime;
							*badop = 0;
						}
						/* Bounce: */
						else
							*badop = 1;
					}
					else if (doesdeop && newtime < chptr->creationtime)
						*badop = 2;
					/* A legal *badop can occur when two people join
					   simultaneously a channel, Allow for 10 min of lag (and
					   thus hacking on channels younger then 10 min) --Run */
					else if (*badop == 0 || chptr->creationtime > (time(NULL) - (time_t) 600)) {
						if (doesop || (!have_ops(chptr) && !PersistantChannel(chptr)))
							chptr->creationtime = newtime;
						*badop = 0;
					}
					break;
				}
			}
			else
				*badop = 0;
		}
	} /* end of while loop for MODE processing */

	if (doesop && newtime == 0 && (IsServer(sptr) || IsULine(cptr, sptr) || sadmin == 1))
		*badop = 2;

	if (*badop >= 2 && (aconf = find_conf_host(cptr->confs, sptr->name, CONF_UWORLD)) != NULL)
		*badop = 4;
	if (*badop >= 2 && (IsULine(cptr, sptr) || sadmin == 1 || hacknotice))
		*badop = 4;

/* Fixes channel hacking.  Problem before was that it didn't bounce
   unless user was deopped by server.  Now modes from *all*
   non-chanops are bounced.  -- Barubary */
	if (!IsServer(sptr) && !IsULine(cptr, sptr) && !is_chan_op(sptr, chptr))
		bounce = 1;
	else
		bounce = (*badop == 1 || *badop == 2) ? 1 : 0;
	if (IsULine(cptr, sptr) || sadmin == 1 || hacknotice)
		bounce = 0;

	whatt = 0;
	for (ip = flags; *ip; ip += 2) {
		if ((*ip & new) && !(*ip & oldm.mode)) {
			if (bounce) {
				if (bwhatt != MODE_DEL) {
					*bmbuf++ = '-';
					bwhatt = MODE_DEL;
				}
				*bmbuf++ = (char)*(ip + 1);
			}
			else {
				if (whatt != MODE_ADD) {
					*mbuf++ = '+';
					whatt = MODE_ADD;
				}
				mode->mode |= *ip;
				*mbuf++ = (char)*(ip + 1);
			}
		}
	}
	for (ip = flags; *ip; ip += 2) {
		if ((*ip & oldm.mode) && !(*ip & new)) {
			if (bounce) {
				if (bwhatt != MODE_ADD) {
					*bmbuf++ = '+';
					bwhatt = MODE_ADD;
				}
				*bmbuf++ = (char)*(ip + 1);
			}
			else {
				if (whatt != MODE_DEL) {
					*mbuf++ = '-';
					whatt = MODE_DEL;
				}
				mode->mode &= ~*ip;
				*mbuf++ = (char)*(ip + 1);
			}
		}
	}

	blen = 0;
	if (limitset && !nusers && mode->limit) {
		if (bounce) {
			if (bwhatt != MODE_ADD) {
				*bmbuf++ = '+';
				bwhatt = MODE_ADD;
			}
			*bmbuf++ = 'l';
			(void) sprintf(numeric, "%-15d", mode->limit);
			if ((cp = strchr(numeric, ' ')) != NULL)
				*cp = 0;
			(void) strcat(bpbuf, numeric);
			blen += strlen(numeric);
			(void) strcat(bpbuf, " ");
		}
		else {
			if (whatt != MODE_DEL) {
				*mbuf++ = '-';
				whatt = MODE_DEL;
			}
			mode->mode &= ~MODE_LIMIT;
			mode->limit = 0;
			*mbuf++ = 'l';
		}
	}

	/* Reconstruct "+bkov" chain. */
	if (opcnt) {
		int i;
		char c = 0;
		u_int32_t prev_whatt = 0;

		for (i = 0; i < opcnt; i++) {
			lp = &chops[i];
			/* make sure we have correct mode change sign */
			if (whatt != (lp->flags & (MODE_ADD | MODE_DEL))) {
				if (lp->flags & MODE_ADD) {
					*mbuf++ = '+';
					prev_whatt = whatt;
					whatt = MODE_ADD;
				}
				else {
					*mbuf++ = '-';
					prev_whatt = whatt;
					whatt = MODE_DEL;
				}
			}

			len = strlen(pbuf);
			/* get c as the mode char and tmp as a pointer to
			 * the parameter for this mode change. */
			switch (lp->flags & MODE_WPARAS) {
				case MODE_CHANOP:
					c = 'o';
					cp = lp->value.cptr->name;
					break;

				case MODE_VOICE:
					c = 'v';
					cp = lp->value.cptr->name;
					break;

				case MODE_BAN:
					/* I made this a bit more user-friendly (tm): nick=nick!*@*
					   nick!user = nick!user@* user@host = *!user@host host.name =
					   *!*@host.name --Run */
					c = 'b';
					/* only process if from a local client */
					cp = MyClient(sptr) ? pretty_mask(lp->value.cp) : lp->value.cp;
					break;

				case MODE_EXC:
					c = 'e';
					cp = MyClient(sptr) ? pretty_mask(lp->value.cp) : lp->value.cp;
					break;

				case MODE_INV:
					c = 'I';
					cp = MyClient(sptr) ? pretty_mask(lp->value.cp) : lp->value.cp;
					break;

				case MODE_KEY:
					c = 'k';
					cp = lp->value.cp;
					break;

				case MODE_LIMIT:
					c = 'l';
					sprintf(numeric, "%-15d", nusers);
					if ((cp = strchr(numeric, ' ')) != NULL)
						*cp = 0;
					cp = numeric;
					break;
			}

			if ((len + strlen(cp) + 12) > (size_t) MODEBUFLEN)
				break;

			switch (lp->flags & MODE_WPARAS) {
				case MODE_KEY:
					if (strlen(cp) > (size_t) KEYLEN)
						*(cp + KEYLEN) = 0;

					if ((whatt == MODE_ADD && (*mode->key == 0 || mycmp(mode->key, cp) != 0)) || (whatt == MODE_DEL && (*mode->key != 0))) {
						if (bounce) {
							if (*mode->key == 0) {
								if (bwhatt != MODE_DEL) {
									*bmbuf++ = '-';
									bwhatt = MODE_DEL;
								}
								(void) strcat(bpbuf, cp);
								blen += strlen(cp);
								(void) strcat(bpbuf, " ");
								blen++;
							}
							else {
								if (bwhatt != MODE_ADD) {
									*bmbuf++ = '+';
									bwhatt = MODE_ADD;
								}
								(void) strcat(bpbuf, mode->key);
								blen += strlen(mode->key);
								(void) strcat(bpbuf, " ");
								blen++;
							}
							*bmbuf++ = c;
							mbuf--;
							if (*mbuf != '+' && *mbuf != '-')
								mbuf++;
							else
								whatt = prev_whatt;
						}
						else {
							*mbuf++ = c;
							(void) strcat(pbuf, cp);
							len += strlen(cp);
							(void) strcat(pbuf, " ");
							len++;
							if (whatt == MODE_ADD)
								strlcpy(mode->key, cp, sizeof(mode->key));
							else
								*mode->key = 0;
						}
					}
					break;

				case MODE_LIMIT:
					if (nusers && nusers != mode->limit) {
						if (bounce) {
							if (mode->limit == 0) {
								if (bwhatt != MODE_DEL) {
									*bmbuf++ = '-';
									bwhatt = MODE_DEL;
								}
							}
							else {
								if (bwhatt != MODE_ADD) {
									*bmbuf++ = '+';
									bwhatt = MODE_ADD;
								}
								(void) sprintf(numeric, "%-15d", mode->limit);
								if ((cp = strchr(numeric, ' ')) != NULL)
									*cp = 0;
								(void) strcat(bpbuf, numeric);
								blen += strlen(numeric);
								(void) strcat(bpbuf, " ");
								blen++;
							}
							*bmbuf++ = c;
							mbuf--;
							if (*mbuf != '+' && *mbuf != '-')
								mbuf++;
							else
								whatt = prev_whatt;
						}
						else {
							*mbuf++ = c;
							(void) strcat(pbuf, cp);
							len += strlen(cp);
							(void) strcat(pbuf, " ");
							len++;
							mode->limit = nusers;
						}
					}
					break;

			case MODE_CHANOP:
			case MODE_VOICE:
				tmp = find_user_link(chptr->members, lp->value.cptr);
				if (lp->flags & MODE_ADD) {
					change = (~tmp->flags) & CHFL_OVERLAP & lp->flags;
					if (change && bounce) {
						if (lp->flags & MODE_CHANOP)
							tmp->flags |= CHFL_DEOPPED;
						if (bwhatt != MODE_DEL) {
							*bmbuf++ = '-';
							bwhatt = MODE_DEL;
						}
						*bmbuf++ = c;
						(void) strcat(bpbuf, lp->value.cptr->name);
						blen += strlen(lp->value.cptr->name);
						(void) strcat(bpbuf, " ");
						blen++;
						change = 0;
					}
					else if (change) {
						tmp->flags |= lp->flags & CHFL_OVERLAP;
						if (lp->flags & MODE_CHANOP) {
							tmp->flags &= ~CHFL_DEOPPED;
							if (IsServer(sptr) || IsULine(cptr, sptr))
								tmp->flags &= ~CHFL_SERVOPOK;
						}
					}
				}
				else {
					change = tmp->flags & CHFL_OVERLAP & lp->flags;
					if (change && bounce) {
						if (lp->flags & MODE_CHANOP)
							tmp->flags &= ~CHFL_DEOPPED;
						if (bwhatt != MODE_ADD) {
							*bmbuf++ = '+';
							bwhatt = MODE_ADD;
						}
						*bmbuf++ = c;
						(void) strcat(bpbuf, lp->value.cptr->name);
						blen += strlen(lp->value.cptr->name);
						(void) strcat(bpbuf, " ");
						blen++;
						change = 0;
					}
					else {
						tmp->flags &= ~change;
						if ((change & MODE_CHANOP) && (IsServer(sptr) || IsULine(cptr, sptr) || sadmin == 1))
							tmp->flags |= CHFL_DEOPPED;
					}
				}
				if (change || *badop == 2 || *badop == 4) {
					*mbuf++ = c;
					(void) strcat(pbuf, cp);
					len += strlen(cp);
					(void) strcat(pbuf, " ");
					len++;
				}
				else {
					mbuf--;
					if (*mbuf != '+' && *mbuf != '-')
						mbuf++;
					else
						whatt = prev_whatt;
				}
				break;

			case MODE_BAN:
				/* prevent forwarding masks on a -f channel */
				if (MyClient(sptr) && (whatt & MODE_ADD)) {
					char *p;
					if ((p = strchr(cp, ',')) != NULL) {
						if (!(new & MODE_FORWARD)) {
							sendto_one(sptr, err_str(ERR_NOFORWARDING), me.name, sptr->name, chptr->chname);
							break;
						}
						else if (!IsChannelName(p + 1)) {
							sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, sptr->name, p + 1);
							break;
						}
					}
				}

				/* Only bans aren't bounced, it makes no sense to bounce last
				   second bans while propagating bans done before the
				   net.rejoin. The reason why I don't bounce net.rejoin bans is
				   because it is too much work to take care of too long strings
				   adding the necessary TS to net.burst bans -- RunLazy We do
				   have to check for *badop==2 now, we don't want HACKs to take
				   effect. */
				/* Not anymore.  They're not bounced from servers/ulines, but
				   they *are* bounced if from a user without chanops. --
				   Barubary */
				if (IsServer(sptr) || IsULine(cptr, sptr) || !bounce) {
					if (*badop != 2 && (((whatt & MODE_ADD) && !add_banid(sptr, chptr, cp)) || ((whatt & MODE_DEL) && !del_banid(chptr, cp)))) {
						*mbuf++ = c;
						(void) strcat(pbuf, cp);
						len += strlen(cp);
						(void) strcat(pbuf, " ");
						len++;
					}
					else {
						/* Borg88 Try to fix bug with redutant bans */
						if (*(mbuf - 1) == '-' || *(mbuf - 1) == '+') {
							whatt = 0;
							*--mbuf = 0;
						}
					}
				}
				else {
					if (whatt & MODE_ADD) {
						if (!find_banid(chptr, cp)) {
							if (bwhatt != MODE_DEL)
								*bmbuf++ = '-';
							*bmbuf++ = c;
							strcat(bpbuf, cp);
							blen += strlen(cp);
							strcat(bpbuf, " ");
							blen++;
						}
					}
					else {
						if (find_banid(chptr, cp)) {
							if (bwhatt != MODE_ADD)
								*bmbuf++ = '+';
							*bmbuf++ = c;
							strcat(bpbuf, cp);
							blen += strlen(cp);
							strcat(bpbuf, " ");
							blen++;
						}
					}
				}
				break;

			case MODE_EXC:
				if (IsServer(sptr) || IsULine(cptr, sptr) || !bounce) {
					if (*badop != 2 && (((whatt & MODE_ADD) && !add_excid(sptr, chptr, cp)) || ((whatt & MODE_DEL) && !del_excid(chptr, cp)))) {
						*mbuf++ = c;
						(void) strcat(pbuf, cp);
						len += strlen(cp);
						(void) strcat(pbuf, " ");
						len++;
					}
					else {
						if (*(mbuf - 1) == '-' || *(mbuf - 1) == '+') {
							whatt = 0;
							*--mbuf = 0;
						}
					}
				}
				else {
					if (whatt & MODE_ADD) {
						if (!find_excid(chptr, cp)) {
							if (bwhatt != MODE_DEL)
								*bmbuf++ = '-';
							*bmbuf++ = c;
							strcat(bpbuf, cp);
							blen += strlen(cp);
							strcat(bpbuf, " ");
							blen++;
						}
					}
					else {
						if (find_excid(chptr, cp)) {
							if (bwhatt != MODE_ADD)
								*bmbuf++ = '+';
							*bmbuf++ = c;
							strcat(bpbuf, cp);
							blen += strlen(cp);
							strcat(bpbuf, " ");
							blen++;
						}
					}
				}
				break;

			case MODE_INV:
				if (IsServer(sptr) || IsULine(cptr, sptr) || !bounce) {
					if (*badop != 2 && (((whatt & MODE_ADD) && !add_invid(sptr, chptr, cp)) || ((whatt & MODE_DEL) && !del_invid(chptr, cp)))) {
						*mbuf++ = c;
						(void) strcat(pbuf, cp);
						len += strlen(cp);
						(void) strcat(pbuf, " ");
						len++;
					}
					else {
						if (*(mbuf - 1) == '-' || *(mbuf - 1) == '+') {
							whatt = 0;
							*--mbuf = 0;
						}
					}
				}
				else {
					if (whatt & MODE_ADD) {
						if (!find_invid(chptr, cp)) {
							if (bwhatt != MODE_DEL)
								*bmbuf++ = '-';
							*bmbuf++ = c;
							strcat(bpbuf, cp);
							blen += strlen(cp);
							strcat(bpbuf, " ");
							blen++;
						}
					}
					else {
						if (find_invid(chptr, cp)) {
							if (bwhatt != MODE_ADD)
								*bmbuf++ = '+';
							*bmbuf++ = c;
							strcat(bpbuf, cp);
							blen += strlen(cp);
							strcat(bpbuf, " ");
							blen++;
						}
					}
				}
				break;
			}
		}	/* for (i=0;i<opcnt;i++) */
	}	/* if(opcnt) */

	*mbuf++ = 0;
	*bmbuf++ = 0;

/* Bounce here */
	if (!hacknotice && *bmodebuf && chptr->creationtime)
		sendto_one(cptr, ":%s %s %s %s %s %lu", me.name, (IsToken(cptr) ? TOK_MODE : MSG_MODE), chptr->chname, bmodebuf, bparambuf, *badop == 2 ? (time_t) 0 : chptr->creationtime);

	return gotts ? 1 : -1;
}

/* Now let _invited_ people join thru bans, +i and +l.
 * Checking if an invite exist could be done only if a block exists,
 * but I'm not too fancy of the complicated structure that'd cause,
 * when optimization will hopefully take care of it. Most of the time
 * a user won't have invites on him anyway. -Donwulff
 */

static int can_join(aClient * sptr, aChannel * chptr, char *key)
{
	struct SLink *lp;

	if ((chptr->mode.mode & MODE_JUPED) && !IsAnOper(sptr))
		return (ERR_UNAVAILRESOURCE);

	if ((chptr->mode.mode & MODE_RGSTRONLY) && !IsARegNick(sptr))
		return (ERR_NEEDREGGEDNICK);

	if (*chptr->mode.key && (BadPtr(key) || mycmp(chptr->mode.key, key)))
		return (ERR_BADCHANNELKEY);

	for (lp = sptr->user->invited; lp; lp = lp->next)
		if (lp->value.chptr == chptr)
			break;

	if ((chptr->mode.mode & MODE_INVITEONLY) && !is_oninv(sptr, chptr) && !lp)
		return (ERR_INVITEONLYCHAN);

	if ((chptr->mode.limit && chptr->users >= chptr->mode.limit) && !lp)
		return (ERR_CHANNELISFULL);

	if (!lp && is_banned(sptr, chptr, 0) /* && !is_oninv(sptr, chptr) */)
		return (ERR_BANNEDFROMCHAN);

	return 0;
}

/* Truncates the name on any non-standard character */
void clean_channelname(char *cn)
{
	while (*cn >= 33 && *cn != ',' && *(u_char*)cn != 160 && *cn != '*' && *cn != '?')
		++cn;

	*cn = '\0';
}

/* Return -1 if mask is present and doesnt match our server name. */
int check_channelmask(aClient * sptr, aClient * cptr, char *chname)
{
	char *s;

	if (*chname == '&' && IsServer(cptr))
		return -1;

	if (!(s = strrchr(chname, ':')))
		return 0;

	++s;

	if (match(s, me.name) || (IsServer(cptr) && match(s, cptr->name))) {
		if (MyClient(sptr))
			sendto_one(sptr, err_str(ERR_BADCHANMASK), me.name, sptr->name, chname);
		return -1;
	}

	return 0;
}

/* Get Channel block for i (and allocate a new channel block, if it didn't exists before). */
aChannel *get_channel(aClient *cptr, char *chname, int flag)
{
	int len;
	aChannel *chptr;

	if (BadPtr(chname))
		return NULL;

	len = strlen(chname);
	if (MyClient(cptr) && len > CHANNELLEN) {
		len = CHANNELLEN;
		*(chname + CHANNELLEN) = '\0';
	}

	if ((chptr = find_channel(chname, (aChannel *) NULL)))
		return chptr;

	if (flag == CREATE) {
		chptr = (aChannel *) MyMalloc(sizeof(aChannel) + len);
		memset(chptr, 0, sizeof(aChannel));
		strlcpy(chptr->chname, chname, len + 1);

		/* The first channel created, pin to the head */
		if (channel)
			channel->prevch = chptr;
		else
			channel = chptr;

		/* Add to the tail also */
		if (chlast != 0) {
			chlast->nextch = chptr;
			chptr->prevch = chlast;
			chptr->nextch = 0;
		}

		chlast = chptr;

		chptr->creationtime = MyClient(cptr) ? time(NULL) : (time_t) 0;
		chptr->mode.mode |= MODE_NOPRIVMSGS;
		add_to_channel_hash_table(chname, chptr);
	}

	return chptr;
}

/*
 * Slight changes in routine, now working somewhat symmetrical:
 *   First try to remove the client & channel pair to avoid duplicates
 *   Second check client & channel invite-list lengths and remove tail
 *   Finally add new invite-links to both client and channel
 * Should U-lined clients have higher limits?   -Donwulff
 */

static void add_invite(cptr, chptr)
aClient *cptr;
aChannel *chptr;
{
	Link *inv, *tmp;

	del_invite(cptr, chptr);
	/*
	 * delete last link in chain if the list is max length
	 */
	if (list_length(cptr->user->invited) >= MAXCHANNELSPERUSER) {
/*             This forgets the channel side of invitation     -Vesa
			   inv = cptr->user->invited;
			   cptr->user->invited = inv->next;
			   free_link(inv);
*/
		for (tmp = cptr->user->invited; tmp->next; tmp = tmp->next);
		del_invite(cptr, tmp->value.chptr);

	}
	/* We get pissy over too many invites per channel as well now, since
	   otherwise mass-inviters could take up some major resources -Donwulff */
	if (list_length(chptr->invites) >= MAXCHANNELSPERUSER) {
		for (tmp = chptr->invites; tmp->next; tmp = tmp->next);
		del_invite(tmp->value.cptr, chptr);
	}
	/*
	 * add client to the beginning of the channel invite list
	 */
	inv = make_link();
	inv->value.cptr = cptr;
	inv->next = chptr->invites;
	chptr->invites = inv;
	/*
	 * add channel to the beginning of the client invite list
	 */
	inv = make_link();
	inv->value.chptr = chptr;
	inv->next = cptr->user->invited;
	cptr->user->invited = inv;
}

/*
 * Delete Invite block from channel invite list and client invite list
 */
void del_invite(cptr, chptr)
aClient *cptr;
aChannel *chptr;
{
	Link **inv, *tmp;

	for (inv = &(chptr->invites); (tmp = *inv); inv = &tmp->next)
		if (tmp->value.cptr == cptr) {
			*inv = tmp->next;
			free_link(tmp);
			break;
		}

	for (inv = &(cptr->user->invited); (tmp = *inv); inv = &tmp->next)
		if (tmp->value.chptr == chptr) {
			*inv = tmp->next;
			free_link(tmp);
			break;
		}
}

/* Subtract one user from channel i (and free channel
   block, if channel became empty). */
static void sub1_from_channel(aChannel *chptr)
{
	Ban *ban;
	Link *lp;

	if (--chptr->users <= 0) {
		/* registered channels persist */
		if (PersistantChannel(chptr)) {
			/* refresh creation time so that the modes we send get
			 * bounced when we join the network back, if there is
			 * an older instance of this channel.
			 */
			chptr->creationtime = time(NULL);

			return;
		}

		/* Now, find all invite links from channel structure */
		while ((lp = chptr->invites))
			del_invite(lp->value.cptr, chptr);

		while (chptr->banlist) {
			ban = chptr->banlist;
			chptr->banlist = ban->next;
			MyFree(ban->banstr);
			MyFree(ban->who);
			free_ban(ban);
		}

		while (chptr->exclist) {
			ban = chptr->exclist;
			chptr->exclist = ban->next;
			MyFree(ban->banstr);
			MyFree(ban->who);
			free_ban(ban);
		}

		while (chptr->invlist) {
			ban = chptr->invlist;
			chptr->invlist = ban->next;
			MyFree(ban->banstr);
			MyFree(ban->who);
			free_ban(ban);
		}

		if (chptr->nextch != 0)
			chptr->nextch->prevch = chptr->prevch;
		if (chptr->prevch != 0)
			chptr->prevch->nextch = chptr->nextch;
		if (channel == chptr)
			channel = chptr->nextch;
		if (chlast == chptr)
			chlast = chptr->prevch;

		(void) del_from_channel_hash_table(chptr->chname, chptr);
		MyFree((char *) chptr);
	}
}

/* m_join
 *      parv[0] = sender prefix
 *      parv[1] = channel
 *      parv[2] = channel password (key)
 */
int m_join(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	static char jbuf[BUFSIZE];
	Link *lp;
	aChannel *chptr;
	char *name, *key = NULL, *p = NULL, *p2 = NULL;
	int syschan, i, flags = 0, zombie = 0;

	if (check_registered_user(sptr))
		return 0;

	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "JOIN");
		return 0;
	}

	*jbuf = '\0';

	/* Rebuild list of channels joined to be the actual result of the */
	/* JOIN.  Note that "JOIN 0" is the destructive problem. */
	for (i = 0, name = strtoken(&p, parv[1], ","); name; name = strtoken(&p, NULL, ",")) {
		/* pathological case only on longest channel name. */
		/* If not dealt with here, causes desynced channel ops */
		/* since ChannelExists() doesn't see the same channel */
		/* as one being joined. cute bug. Oct 11 1997, Dianora/comstud */
		/* Copied from Dianora's "hybrid 5" ircd. */

		if (strlen(name) > CHANNELLEN) /* same thing is done in get_channel() */
			name[CHANNELLEN] = '\0';

		if (MyConnect(sptr)) {
			aChannel *tch;

			clean_channelname(name);

			/* correct channel name on local joins -- hex */
			if ((tch = find_channel(name, NullChn))) {
				name = tch->chname;

				/* forward joins */
				if (tch->mode.mode & MODE_FORWARD) {
					Ban *ban;
					char *target;

					if ((ban = is_banned(sptr, tch, 0)) && (target = strchr(ban->banstr, ',')) && !is_oninv(sptr, tch)) {
						/* skip the comma */
						++target;

						/* correct the name again */
						if ((tch = find_channel(target, NullChn)))
							target = tch->chname;

						sendto_one(sptr, rpl_str(RPL_FORWARD), me.name, sptr->name, name, target);
						name = target;
					}
				}
			}
		}

		if (check_channelmask(sptr, cptr, name) == -1)
			continue;

		if (*name == '&' && !MyConnect(sptr))
			continue;

		if (*name == '0' && !atoi(name)) {
			strcpy(jbuf, "0");
			i = 1;
			continue;
		}
		else if (!IsChannelName(name)) {
			if (MyClient(sptr))
				sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], name);
			continue;
		}
		else if (*name == '&' && !IsAnOper(sptr) && !strncasecmp(name, "&ircd.", 6)) {
			sendto_one(sptr, ":%s 477 %s %s :You need to be an IRC Operator to join that channel", me.name, parv[0], name);
			continue;
		}

		if (*jbuf)
			strcat(jbuf, ",");
		strncat(jbuf, name, sizeof(jbuf) - i - 1);
		i += strlen(name) + 1;
	}

	(void) strcpy(parv[1], jbuf);

	p = NULL;
	if (parv[2])
		key = strtoken(&p2, parv[2], ",");

	parv[2] = NULL; /* for m_names call later, parv[parc] must == NULL */

	for (name = strtoken(&p, jbuf, ","); name; key = (key) ? strtoken(&p2, NULL, ",") : NULL, name = strtoken(&p, NULL, ",")) {
		/* JOIN 0 sends out a part for all channels a user has joined. */
		if (*name == '0' && !atoi(name)) {
			while ((lp = sptr->user->channel)) {
				chptr = lp->value.chptr;
				sendto_channel_butserv(chptr, sptr, PartFmt, parv[0], chptr->chname);
				remove_user_from_channel(sptr, chptr);
			}
			sendto_serv_butone(cptr, ":%s JOIN 0", parv[0]);
			continue;
		}

		syschan = !strncasecmp(name, "&ircd.", 6) ? 1 : 0;

		if (MyConnect(sptr)) {
			/* local client is first to enter previously nonexistant */
			/* channel so make them (rightfully) the Channel */
			/* Operator. */
			flags = !IsModelessChannel(name) && !ChannelExists(name) && !syschan ? CHFL_CHANOP : CHFL_DEOPPED;

			if (sptr->user->joined >= MAXCHANNELSPERUSER) {
				sendto_one(sptr, err_str(ERR_TOOMANYCHANNELS), me.name, parv[0], name);
				return 0;
			}
		}

		chptr = get_channel(sptr, name, CREATE);

		if (syschan)
			chptr->mode.mode |= MODE_TOPICLIMIT|MODE_SECRET;

		if (chptr && (lp = find_user_link(chptr->members, sptr))) {
			if (lp->flags & CHFL_ZOMBIE) {
				zombie = 1;
				flags = lp->flags & (CHFL_DEOPPED | CHFL_SERVOPOK);
				remove_user_from_channel(sptr, chptr);
				chptr = get_channel(sptr, name, CREATE);
			}
			else
				continue;
		}
		if (!zombie) {
			if (!MyConnect(sptr))
				flags = CHFL_DEOPPED;
			if (sptr->flags & FLAGS_TS8)
				flags |= CHFL_SERVOPOK;
		}
		if (!chptr || (MyConnect(sptr) && (i = can_join(sptr, chptr, key)))) {
			sendto_one(sptr, err_str(i), me.name, parv[0], name);
			continue;
		}

		/* Complete user entry to the new channel (if any) */
		add_user_to_channel(chptr, sptr, flags);

		/* notify all other users on the new channel */
		sendto_channel_butserv(chptr, sptr, ":%s JOIN :%s", parv[0], name);
		sendto_match_servs(chptr, cptr, ":%s JOIN :%s", parv[0], name);

		if (MyClient(sptr)) {
			/* Make a (temporal) creationtime, if someone joins */
			/* during a net.reconnect : between remote join and */
			/* the mode with TS. --Run */
			if (chptr->creationtime == 0) {
				chptr->creationtime = time(NULL);
				sendto_match_servs(chptr, cptr, ":%s MODE %s + %lu", me.name, name, chptr->creationtime);
			}

			del_invite(sptr, chptr);

			if (syschan) {
				strcpy(chptr->topic, "IRC Daemon debug channel");
			}

			if (flags & CHFL_CHANOP)
				sendto_match_servs(chptr, cptr, ":%s MODE %s +o %s %lu", me.name, name, parv[0], chptr->creationtime);

			if (chptr->topic[0] != '\0') {
				sendto_one(sptr, rpl_str(RPL_TOPIC), me.name, parv[0], name, chptr->topic);
				sendto_one(sptr, rpl_str(RPL_TOPICWHOTIME), me.name, parv[0], name, chptr->topic_nick, chptr->topic_time);
			}

			parv[1] = name;
			(void) m_names(cptr, sptr, 2, parv);
		}
	}
	return 0;
}


/* m_kick */
/*     parv[0] = sender prefix */
/*     parv[1] = channel */
/*     parv[2] = client to kick */
/*     parv[3] = kick comment */
int m_kick(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	aClient *who;
	aChannel *chptr;
	int chasing = 0, count;
	char *comment, *name, *p = NULL, *user, *p2 = NULL;
	struct SLink *lp, *lp2;

	if (check_registered(sptr))
		return 0;

	sptr->flags &= ~FLAGS_TS8;

	if (parc < 3 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "KICK");
		return 0;
	}

	if (IsServer(sptr) && !IsULine(cptr, sptr))
		sendto_ops("HACK: KICK from %s for %s %s", parv[0], parv[1], parv[2]);

	comment = (BadPtr(parv[3])) ? parv[0] : parv[3];
	if (strlen(comment) > (size_t) TOPICLEN)
		comment[TOPICLEN] = '\0';

	*nickbuf = *buf = '\0';

	for (count = 0; (name = strtoken(&p, parv[1], ",")) && count < 4; parv[1] = NULL) {
		chptr = get_channel(sptr, name, !CREATE);
		if (!chptr) {
			sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], name);
			continue;
		}
		if (check_channelmask(sptr, cptr, name))
			continue;
		if (!IsServer(cptr) && !IsULine(cptr, sptr) && !is_chan_op(sptr, chptr)) {
			sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0], chptr->chname);
			continue;
		}

		lp2 = find_user_link(chptr->members, sptr);
		for (; (user = strtoken(&p2, parv[2], ",")); parv[2] = NULL) {
			if (!(who = find_chasing(sptr, user, &chasing)))
				continue;	/* No such user left! */

			/* services ain't kickable */
			if (IsULine(who, who)) {
				sendto_one(sptr, ":%s NOTICE %s :Cannot kick services.", me.name, who->name);
				continue;
			}

			if (((lp = find_user_link(chptr->members, who)) && !(lp->flags & CHFL_ZOMBIE)) || IsServer(sptr)) {
				/* Bounce all KICKs from a non-chanop unless it
				 * would cause a fake direction -- Barubary */
				if ((who->from != cptr) && !IsServer(sptr) && !IsULine(cptr, sptr) && !is_chan_op(sptr, chptr)) {
					/* Bounce here: */
					/* cptr must be a server (or cptr==sptr and */
					/* sptr->flags can't have DEOPPED set */
					/* when CHANOP is set). */
					sendto_one(cptr, ":%s JOIN %s", who->name, name);
					/* Slight change here - the MODE is sent only */
					/* once even if overlapping -- Barubary */
					if (lp->flags & CHFL_OVERLAP) {
						char *temp = buf;

						*temp++ = '+';
						if (lp->flags & CHFL_CHANOP)
							*temp++ = 'o';
						if (lp->flags & CHFL_VOICE)
							*temp++ = 'v';
						*temp = 0;
						sendto_one(cptr, ":%s MODE %s %s %s %s %lu", me.name, chptr->chname, buf, who->name, ((lp->flags & CHFL_OVERLAP) == CHFL_OVERLAP) ? who->name : "", chptr->creationtime);
					}
				}
				else {
					if (lp)
						sendto_channel_butserv(chptr, sptr, ":%s KICK %s %s :%s", parv[0], name, who->name, comment);
					sendto_match_servs(chptr, cptr, ":%s KICK %s %s :%s", parv[0], name, who->name, comment);
					/* Finally, zombies totally removed -- Barubary */
					if (lp) {
						if (MyClient(who)) {
							sendto_match_servs(chptr, NULL, PartFmt, who->name, name);
							remove_user_from_channel(who, chptr);
						}
						else {
							remove_user_from_channel(who, chptr);
						}
					}
				}
			}
			else if (MyClient(sptr))
				sendto_one(sptr, err_str(ERR_USERNOTINCHANNEL), me.name, parv[0], user, name);
			if (!IsServer(cptr) || !IsULine(cptr, sptr))
				count++;
		}	/* loop on parv[2] */
		if (!IsServer(cptr) || !IsULine(cptr, sptr))
			break;
	}	/* loop on parv[1] */
	return (0);
}

int count_channels(aClient *sptr)
{
	aChannel *chptr;
	int count = 0;

	for (chptr = channel; chptr; chptr = chptr->nextch) {
		/* hide empty persistant channels from the masses */
		if (!chptr->members)
			continue;

		if (SecretChannel(chptr)) {
			if (IsAnOper(sptr))
				++count;
		}
		else
			++count;
	}

	return count;
}

/*
** m_topic
**     parv[0] = sender prefix
**     parv[1] = topic text
**
**     For servers using TS: (Lefler)
**     parv[0] = sender prefix
**     parv[1] = channel list
**     parv[2] = topic nickname
**     parv[3] = topic time
**     parv[last] = topic text
*/
int m_topic(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	aChannel *chptr = NullChn;
	char *topic = NULL, *name, *p = NULL, *tnick = NULL;
	time_t ttime = 0;

	if (check_registered(sptr))
		return 0;

	if (parc < 2) {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "TOPIC");
		return 0;
	}

	if (parc > 2) {
		topic = parv[parc - 1];
		if (parc > 4) {
			tnick = parv[2];
			ttime = atol(parv[3]);
		}
	}

	for (; (name = strtoken(&p, parv[1], ",")); parv[1] = NULL) {
		chptr = NULL;

		if (!IsChannelName(name) || !(chptr = find_channel(name, NullChn))) {
			if (IsPerson(sptr))
				sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL),
					me.name, parv[0], name);
			continue;
		}

		if (IsPerson(sptr) && !IsULine(sptr, sptr) && !IsMember(sptr, chptr)) {
			sendto_one(sptr, err_str(ERR_NOTONCHANNEL),
				me.name, parv[0], name);
			continue;
		}

		if (check_channelmask(sptr, cptr, name))
			continue;

		/* only asking for topic */
		if (topic == NULL) {
			if (chptr->topic[0] == '\0')
				sendto_one(sptr, rpl_str(RPL_NOTOPIC), me.name,
				parv[0], chptr->chname);
			else {
				sendto_one(sptr, rpl_str(RPL_TOPIC), me.name,
					parv[0], chptr->chname, chptr->topic);
				sendto_one(sptr, rpl_str(RPL_TOPICWHOTIME),
					me.name, parv[0], chptr->chname,
					chptr->topic_nick, chptr->topic_time);
			}
		}

		else if (ttime && topic && (IsServer(sptr) || IsULine(cptr, sptr))) {
			if (!chptr->topic_time || ttime > chptr->topic_time) {
				const char *denymsg = can_send(sptr, chptr, topic);
				if (denymsg != NULL && !IsULine(cptr, sptr))
					sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN), me.name, parv[0], sptr->name, denymsg ? denymsg : "no reason");
				else {
					/* setting a topic */
					strlcpy(chptr->topic, topic, sizeof(chptr->topic));
					strcpy(chptr->topic_nick, tnick);
					chptr->topic_time = ttime;
					sendto_match_servs(chptr, cptr, ":%s TOPIC %s %s %lu :%s", parv[0], chptr->chname, chptr->topic_nick, chptr->topic_time, chptr->topic);
					sendto_channel_butserv(chptr, sptr, ":%s TOPIC %s :%s", parv[0], chptr->chname, chptr->topic);
				}
			}
		}
		else if ((((chptr->mode.mode & MODE_TOPICLIMIT) == 0 || is_chan_op(sptr, chptr)) || IsULine(cptr, sptr)) && topic) {
			/* setting a topic */
			strlcpy(chptr->topic, topic, sizeof(chptr->topic));
			strcpy(chptr->topic_nick, sptr->name);
			if (ttime && IsServer(cptr))
				chptr->topic_time = ttime;
			else
				chptr->topic_time = time(NULL);
			sendto_match_servs(chptr, cptr, ":%s TOPIC %s %s %lu :%s", parv[0], chptr->chname, parv[0], chptr->topic_time, chptr->topic);
			sendto_channel_butserv(chptr, sptr, ":%s TOPIC %s :%s", parv[0], chptr->chname, chptr->topic);
		}
		else
			sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED),
				me.name, parv[0], chptr->chname);
	}

	return 0;
}

/*
** m_invite
**     parv[0] - sender prefix
**     parv[1] - user to invite
**     parv[2] - channel number
*/
int m_invite(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aClient *acptr;
	aChannel *chptr;

	if (check_registered_user(sptr))
		return 0;

	if (parc < 3 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "INVITE");
		return -1;
	}

	if (!(acptr = find_person(parv[1], (aClient *) NULL))) {
		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], parv[1]);
		return 0;
	}
	if (MyConnect(sptr))
		clean_channelname(parv[2]);
	if (check_channelmask(sptr, cptr, parv[2]))
		return 0;
	if (!(chptr = find_channel(parv[2], NullChn))) {

		sendto_prefix_one(acptr, sptr, ":%s INVITE %s :%s", parv[0], parv[1], parv[2]);
		return 0;
	}

	if (chptr && !IsMember(sptr, chptr) && !IsULine(cptr, sptr)) {
		sendto_one(sptr, err_str(ERR_NOTONCHANNEL), me.name, parv[0], parv[2]);
		return -1;
	}

	if (IsMember(acptr, chptr)) {
		sendto_one(sptr, err_str(ERR_USERONCHANNEL), me.name, parv[0], parv[1], parv[2]);
		return 0;
	}
	if (chptr && (chptr->mode.mode & MODE_INVITEONLY)) {
		if (!is_chan_op(sptr, chptr) && !IsULine(cptr, sptr)) {
			sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0], chptr->chname);
			return -1;
		}
		else if (!IsMember(sptr, chptr) && !IsULine(cptr, sptr)) {
			sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0], ((chptr) ? (chptr->chname) : parv[2]));
			return -1;
		}
	}

	if (MyConnect(sptr)) {
		if (check_for_target_limit(sptr, acptr, acptr->name))
			return 0;

		sendto_one(sptr, rpl_str(RPL_INVITING), me.name, parv[0], acptr->name, ((chptr) ? (chptr->chname) : parv[2]));
		if (IsAway(acptr))
			sendto_one(sptr, rpl_str(RPL_AWAY), me.name, parv[0], acptr->name, MyClient(acptr) ? acptr->away : "Gone.");
	}

	/* Note: is_banned() here will cause some extra CPU load, and we're really
	   only relying on the existence of the limit because we could momentarily
	   have less people on channel. */
	if (MyConnect(acptr)) {
		if (chptr && sptr->user && (is_banned(acptr, chptr, 0) || (chptr->mode.mode & MODE_INVITEONLY) || chptr->mode.limit) && (is_chan_op(sptr, chptr) || IsULine(cptr, sptr))) {
			add_invite(acptr, chptr);
			sendto_channelops_butone(NULL, &me, chptr, ":%s NOTICE @%s :%s invited %s into the channel.", me.name, chptr->chname, sptr->name, acptr->name);
		}
	}

	sendto_prefix_one(acptr, sptr, ":%s INVITE %s :%s", parv[0], acptr->name, ((chptr) ? (chptr->chname) : parv[2]));
	return 0;
}

/************************************************************************
 * m_names() - Added by Jto 27 Apr 1989
 ************************************************************************/

/*
** m_names
**     parv[0] = sender prefix
**     parv[1] = channel
*/

int m_names(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	aChannel *chptr, *ch2ptr = NULL;
	aClient *c2ptr;
	Link *lp;
	int idx, flag, len, mlen, x;
	char *s, *para = parc > 1 ? parv[1] : NULL;

	if (check_registered(sptr))
		return 0;

	if (parc > 1 && hunt_server(cptr, sptr, ":%s NAMES %s %s", 2, parc, parv))
		return 0;

	mlen = strlen(me.name) + 10;

	/* Only 10 requests per NAMES allowed remotely -- Barubary */
	if (!BadPtr(para) && (cptr != sptr))
		for (x = 0, s = para; *s; s++)
			if (*s == ',' && ++x == 10) {
				*s = 0;
				break;
			}

	if (BadPtr(para) && IsServer(cptr)) {
		sendto_one(cptr, rpl_str(RPL_ENDOFNAMES), me.name, parv[0], "*");
		return 0;
	}

	if (!BadPtr(para)) {
		s = strchr(para, ',');
		if (s) {
			parv[1] = ++s;
			(void) m_names(cptr, sptr, parc, parv);
		}
		if (MyConnect(sptr))
			clean_channelname(para);
		ch2ptr = find_channel(para, (aChannel *) NULL);
	}

	*buf = '\0';

	/* First, do all visible channels (public and the one user self is) */

	for (chptr = channel; chptr; chptr = chptr->nextch) {
		if ((chptr != ch2ptr) && !BadPtr(para))
			continue;	/* wanted a specific channel */
		if (!MyConnect(sptr) && BadPtr(para))
			continue;
		if (!ShowChannel(sptr, chptr))
			continue;	/* users on this are not listed */

		/* Find users on same channel (defined by chptr) */

		(void) strcpy(buf, "* ");
		len = strlen(chptr->chname);
		(void) strcpy(buf + 2, chptr->chname);
		(void) strcpy(buf + 2 + len, " :");

		*buf = (char)(SecretChannel(chptr) ? '@' : '=');

		idx = len + 4;
		flag = 1;
		for (lp = chptr->members; lp; lp = lp->next) {
			c2ptr = lp->value.cptr;
			if (sptr != c2ptr && (IsInvisible(c2ptr) && !IsAnOper(sptr)) && !IsMember(sptr, chptr))
				continue;

			if (lp->flags & CHFL_ZOMBIE) {
				if (lp->value.cptr != sptr)
					continue;
				else
					(void) strcat(buf, "!");
			}
			else if (lp->flags & CHFL_CHANOP)
				(void) strcat(buf, "@");
			else if (lp->flags & CHFL_VOICE)
				(void) strcat(buf, "+");

			(void) strncat(buf, c2ptr->name, NICKLEN);
			idx += strlen(c2ptr->name) + 1;
			flag = 1;
			(void) strcat(buf, " ");

			if (mlen + idx + NICKLEN > BUFSIZE - 2) {
				sendto_one(sptr, rpl_str(RPL_NAMREPLY), me.name, parv[0], buf);
				(void) strncpy(buf, "* ", 3);
				(void) strncpy(buf + 2, chptr->chname, len + 1);
				(void) strcat(buf, " :");
				*buf = (char)(SecretChannel(chptr) ? '@' : '=');
				idx = len + 4;
				flag = 0;
			}
		}
		if (flag)
			sendto_one(sptr, rpl_str(RPL_NAMREPLY), me.name, parv[0], buf);
	}

	if (!BadPtr(para)) {
		sendto_one(sptr, rpl_str(RPL_ENDOFNAMES), me.name, parv[0], para);
		return (1);
	}

	/* Second, do all non-public, non-secret channels in one big sweep */

	(void) strncpy(buf, "* * :", 6);
	idx = 5;
	flag = 0;
	for (c2ptr = client; c2ptr; c2ptr = c2ptr->next) {
		int showflag = 0, secret = 0;
		aChannel *ch3ptr;

		if (!IsPerson(c2ptr))
			continue;

		if (sptr != c2ptr && IsInvisible(c2ptr))
			continue;

		lp = c2ptr->user->channel;
		/* dont show a client if they are on a secret channel or they are on a
		   channel sptr is on since they have already been show earlier.
		   -avalon */
		while (lp) {
			ch3ptr = lp->value.chptr;

			if (!SecretChannel(ch3ptr) || IsMember(sptr, ch3ptr))
				showflag = 1;

			if (SecretChannel(ch3ptr))
				secret = 1;

			lp = lp->next;
		}
		if (showflag) /* have we already shown them ? */
			continue;
		if (secret)	/* on any secret channels ? */
			continue;
		(void) strncat(buf, c2ptr->name, NICKLEN);
		idx += strlen(c2ptr->name) + 1;
		(void) strcat(buf, " ");
		flag = 1;
		if (mlen + idx + NICKLEN > BUFSIZE - 2) {
			sendto_one(sptr, rpl_str(RPL_NAMREPLY), me.name, parv[0], buf);
			(void) strncpy(buf, "* * :", 6);
			idx = 5;
			flag = 0;
		}
	}
	if (flag)
		sendto_one(sptr, rpl_str(RPL_NAMREPLY), me.name, parv[0], buf);
	sendto_one(sptr, rpl_str(RPL_ENDOFNAMES), me.name, parv[0], "*");
	return (1);
}

void send_user_joins(aClient * cptr, aClient * user)
{
	Link *lp;
	aChannel *chptr;
	int cnt = 0, len = 0, clen;
	char *mask;

	sprintf(buf, ":%s %s ", user->name, (IsToken(cptr) ? TOK_JOIN : MSG_JOIN));
	len = strlen(buf);

	for (lp = user->user->channel; lp; lp = lp->next) {
		chptr = lp->value.chptr;
		if ((mask = strchr(chptr->chname, ':')))
			if (match(++mask, cptr->name))
				continue;
		if (*chptr->chname == '&')
			continue;
		clen = strlen(chptr->chname);
		if (clen + 1 + len > BUFSIZE - 3) {
			if (cnt) {
				buf[len - 1] = '\0';
				sendto_one(cptr, "%s", buf);
			}
			sprintf(buf, ":%s %s ", user->name, (IsToken(cptr) ? TOK_JOIN : MSG_JOIN));
			len = strlen(buf);
			cnt = 0;
		}
		(void) strcpy(buf + len, chptr->chname);
		cnt++;
		len += clen;
		if (lp->next) {
			len++;
			(void) strcat(buf, ",");
		}
	}
	if (*buf && cnt)
		sendto_one(cptr, "%s", buf);

	return;
}
