/************************************************************************
 * IRC - Internet Relay Chat, ircd/s_auth.c
 * Copyright (C) 1992 Darren Reed
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: s_auth.c,v 1.9 2002/11/18 12:44:08 hex Exp $
 */

#include <string.h>

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "res.h"
#include "numeric.h"

#ifndef _WIN32
# include <sys/socket.h>
# include <sys/file.h>
# include <sys/ioctl.h>
# include <unistd.h>
# if defined(__hpux)
#  include "inet.h"
# endif
#else
# include <io.h>
#endif

#include <fcntl.h>
#include "sock.h" /* BSD4.2 needs FD_ZERO here */
#include "h.h"
#include "ircdlog.h"

/*
 * send_authports
 *
 * Send the ident server a query giving "theirport , ourport".
 * The write is only attempted *once* so it is deemed to be a fail if the
 * entire write doesn't write all the data given.  This shouldnt be a
 * problem since the socket should have a write buffer far greater than
 * this message to store it in should problems arise. -avalon
 */
void send_authports(aClient * cptr)
{
	struct sockaddr_in us, them;
	char authbuf[32];
	int ulen, tlen;

	Debug((DEBUG_NOTICE, "write_authports(%x) fd %d authfd %d stat %d", cptr, cptr->fd, cptr->authfd, cptr->status));
	tlen = ulen = sizeof(us);
	if (getsockname(cptr->fd, (struct sockaddr *) &us, &ulen) || getpeername(cptr->fd, (struct sockaddr *) &them, &tlen)) {
		putlog(LOG_ERR, "auth get{sock,peer}name error for %s:%m",
			get_client_name(cptr, TRUE));
		goto authsenderr;
	}

	(void) sprintf(authbuf, "%u , %u\r\n", (unsigned int) ntohs(them.sin_port), (unsigned int) ntohs(us.sin_port));

	Debug((DEBUG_SEND, "sending [%s] to auth port %s.113", authbuf, inetntoa((char *) &them.sin_addr)));
#ifndef _WIN32
	if (write(cptr->authfd, authbuf, strlen(authbuf)) != (int)strlen(authbuf))
#else
	if (send(cptr->authfd, authbuf, strlen(authbuf), 0) != (int) strlen(authbuf))
#endif
	{
authsenderr:
		ircstp->is_abad++;
#ifndef _WIN32
		(void) close(cptr->authfd);
#else
		(void) closesocket(cptr->authfd);
#endif
		if (cptr->authfd == highest_fd)
			while (!local[highest_fd])
				highest_fd--;
		cptr->authfd = -1;
		cptr->flags &= ~FLAGS_AUTH;
		if (show_connect_info)
			send(cptr->fd, REPORT_FAIL_ID, strlen(REPORT_FAIL_ID), 0);
		if (!DoingDNS(cptr))
			SetAccess(cptr);
	}
	cptr->flags &= ~FLAGS_WRAUTH;
	return;
}

/*
 * read_authports
 *
 * read the reply (if any) from the ident server we connected to.
 * The actual read processijng here is pretty weak - no handling of the reply
 * if it is fragmented by IP.
 */
void read_authports(aClient * cptr)
{
	char *s, *t;
	int len;
	char ruser[USERLEN + 1], system[8];
	u_short remp = 0, locp = 0;

	*system = *ruser = '\0';
	Debug((DEBUG_NOTICE, "read_authports(%x) fd %d authfd %d stat %d", cptr, cptr->fd, cptr->authfd, cptr->status));
	/* 
	 * Nasty.  Cant allow any other reads from client fd while we're
	 * waiting on the authfd to return a full valid string.  Use the
	 * client's input buffer to buffer the authd reply.
	 * Oh. this is needed because an authd reply may come back in more
	 * than 1 read! -avalon
	 */
#ifndef _WIN32
	if ((len = read(cptr->authfd, cptr->buffer + cptr->count, sizeof(cptr->buffer) - 1 - cptr->count)) >= 0)
#else
	if ((len = recv(cptr->authfd, cptr->buffer + cptr->count, sizeof(cptr->buffer) - 1 - cptr->count, 0)) >= 0)
#endif
	{
		cptr->count += len;
		cptr->buffer[cptr->count] = '\0';
	}

	cptr->lasttime = time(NULL);
	if ((len > 0) && (cptr->count != (sizeof(cptr->buffer) - 1)) && (sscanf(cptr->buffer, "%hd , %hd : USERID : %*[^:]: %10s", &remp, &locp, ruser) == 3)) {
		s = strrchr(cptr->buffer, ':');
		*s++ = '\0';
		for (t = (strrchr(cptr->buffer, ':') + 1); *t; t++)
			if (!isspace(*t))
				break;
		strncpyzt(system, t, sizeof(system));
		for (t = ruser; *s && *s != '@' && (t < ruser + sizeof(ruser)); s++)
			if (!isspace(*s) && *s != ':')
				*t++ = *s;
		*t = '\0';
		Debug((DEBUG_INFO, "auth reply ok [%s] [%s]", system, ruser));
	}
	else if (len != 0) {
		if (!strchr(cptr->buffer, '\n') && !strchr(cptr->buffer, '\r'))
			return;
		Debug((DEBUG_ERROR, "local %d remote %d", locp, remp));
		Debug((DEBUG_ERROR, "bad auth reply in [%s]", cptr->buffer));
		*ruser = '\0';
	}
#ifndef _WIN32
	(void) close(cptr->authfd);
#else
	(void) closesocket(cptr->authfd);
#endif
	if (cptr->authfd == highest_fd)
		while (!local[highest_fd])
			highest_fd--;
	cptr->count = 0;
	cptr->authfd = -1;
	ClearAuth(cptr);
	if (!DoingDNS(cptr))
		SetAccess(cptr);
	if (len > 0) {
		Debug((DEBUG_INFO, "ident reply: [%s]", cptr->buffer));
	}

	if (show_connect_info)
		send(cptr->fd, REPORT_FIN_ID, strlen(REPORT_FIN_ID), 0);

	if (!locp || !remp || !*ruser) {
		ircstp->is_abad++;
		return;
	}
	ircstp->is_asuc++;
	strncpyzt(cptr->username, ruser, USERLEN + 1);
	if (strncmp(system, "OTHER", 5))
		cptr->flags |= FLAGS_GOTID;
	Debug((DEBUG_INFO, "got username [%s]", ruser));
	return;
}
